from typing import Any, Callable, List

import traceback
import inspect
import sys

from threading import RLock
from functools import wraps

from bgez.core.abstract import Abstract, abstractmethod
from bgez.core.events import Subject
from bgez import logger, loop

__all__ = [
    'BasePromise', 'Promise',
    'BaseDeferred', 'Deferred',
    'UNRESOLVED',
]

UNRESOLVED = object()

class BasePromise(Abstract):
    '''
    Base class defining the API for promise objects:
     - Handle resolution on `.then(lambda value: ...)`.
     - Handle rejection on `.catch(lambda exc: ...)`.

    Resolving a Promise to a Promise will not return a Promise<Promise>.
    Instead the inner promise will be wrapped by the outer promise,
    and relay its own resolution to the others.
    '''

    UNRESOLVED = UNRESOLVED

    @abstractmethod
    def then(self, callback: Callable) -> None:
        '''Callback to execute when the promise gets resolved.'''

    @abstractmethod
    def catch(self, callback: Callable) -> None:
        '''Callback to execute when the promise gets rejected.'''

    @abstractmethod
    def peek(self, fallback: Any = UNRESOLVED) -> Any:
        '''Try to get the current value of the promise.'''

class BaseDeferred(Abstract):
    '''
    Base class defining the API for deferred objects.
    Inherits from `Promise`.
    - Resolve with `resolve(value)`.
    - Reject with `reject(value)`.
    '''

    @abstractmethod
    def resolve(self, value: Any = None) -> None:
        '''Resolves this deferred instance.'''

    @abstractmethod
    def reject(self, exc: Exception) -> None:
        '''Rejects this deferred instance.'''

class MetaPromise(type):
    '''Defines class methods for the `Promise` type'''

    def resolved(cls, value: Any, **kargs) -> BasePromise:
        return cls(lambda resolve, _: resolve(value), **kargs)

    def rejected(cls, exc: Exception, **kargs) -> BasePromise:
        return cls(lambda _, reject: reject(exc), **kargs)

    def all(cls, promises: List[BasePromise]) -> BasePromise:
        def resolver(resolve, reject):
            results = [Promise.UNRESOLVED for _ in promises]
            finished, total = 0, len(results)

            for i, promise in enumerate(promises):

                def factory(index):
                    def callback(value):
                        nonlocal finished
                        results[index] = value
                        finished += 1
                        if finished == total:
                            resolve(results)
                    return callback

                promise.then(factory(i))
                promise.catch(reject)

        return cls(resolver)

    def race(cls, promises: List[BasePromise]) -> BasePromise:
        def resolver(resolve, reject):
            for promise in promises:
                promise.then(resolve)
                promise.catch(reject)
        return cls(resolver)

class Promise(BasePromise, metaclass=MetaPromise):

    def __init__(self, resolver: Callable):
        self.__loop = loop.get()
        self.__lock = RLock()

        self.__onresolve = Subject()
        self.__onreject = Subject()

        self.__value = UNRESOLVED
        self.__exc = None

        self.__init_stack = traceback.format_stack(inspect.stack()[1].frame, 3)
        self.__handled = False

        resolver(self.__resolve, self.__reject)

    def __resolve(self, value: Any = None) -> None:
        if self.__pending():
            if value is UNRESOLVED:
                raise ValueError('cannot resolve to `UNRESOLVED`')

            # Promise-Chaining: do not resolve before the other one does.
            if isinstance(value, BasePromise):
                value.then(self.__resolve)
                value.catch(self.__reject)
                return

            self.__value = value
            self.__loop.immediate(self.__do_resolve, value)

    def __reject(self, exc: Exception) -> None:
        if self.__pending():
            if not isinstance(exc, Exception):
                raise ValueError(f'{exc} is not an exception')

            self.__exc = exc
            self.__loop.immediate(self.__do_reject, exc)

    def __do_resolve(self, v):
        onresolve = self.__onresolve
        onresolve.notify(v, catch=False)
        onresolve.clear()

    def __do_reject(self, e: Exception):
        onreject = self.__onreject

        if not self.__handled:
            message = '\n'.join(e.args)
            init_stack = ''.join(self.__init_stack)
            e.args = (f'{message}\n\nTraceback (task creation):\n{init_stack}',)
            raise e

        onreject.notify(e, catch=False)
        onreject.clear()

    def __pending(self) -> bool:
        with self.__lock:
            return (self.__value is UNRESOLVED) and (self.__exc is None)

    def then(self, callback: Callable) -> None:
        def resolver(resolve, reject):

            @wraps(callback)
            def wrapper(*args, **kargs):
                try:
                    value = callback(*args, **kargs)
                    resolve(value)
                except Exception as exc:
                    reject(exc)

            if self.__pending():
                self.__onresolve.subscribe(wrapper)
            elif self.__value is not UNRESOLVED:
                self.__loop.immediate(wrapper, self.__value)

        return Promise(resolver)

    def catch(self, callback: Callable) -> None:
        def resolver(resolve, reject):

            @wraps(callback)
            def wrapper(*args, **kargs):
                try:
                    value = callback(*args, **kargs)
                    resolve(value)
                except Exception as exc:
                    reject(exc)

            if self.__pending():
                self.__onreject.subscribe(wrapper)
            elif self.__exc is not None:
                self.__loop.immediate(wrapper, self.__exc)
            self.__handled = True

        return Promise(resolver)

    def peek(self, fallback: Any = UNRESOLVED) -> Any:
        if self.__pending():
            return fallback
        elif self.__exc is not None:
            raise self.__exc # pylint: disable=E0702
        return self.__value

    def __await_callback(self, executor):
        self.catch(executor.throw)
        self.then(executor.send)

    def __await__(self):
        if self.__pending():
            return (yield self.__await_callback)
        return self.peek()

class MetaDeferred(type):
    '''Defines class methods for the `Deferred` type'''

    def resolved(cls, value: Any) -> BaseDeferred:
        deferred = cls()
        deferred.resolve(value)
        return deferred

    def rejected(cls, exc: Exception) -> BaseDeferred:
        deferred = cls()
        deferred.reject(exc)
        return deferred

class Deferred(BaseDeferred, metaclass=MetaDeferred):

    def __extract(self, resolve, reject):
        self.resolve = resolve
        self.reject = reject

    def __init__(self):
        self.promise = Promise(self.__extract)

    def __await__(self):
        return self.promise.__await__()
