from typing import Any, Union, Callable, Hashable, Iterable

from collections import OrderedDict, defaultdict, deque
from functools import partial

from bgez.core.abstract import Abstract, abstractmethod
from bgez.core import decorator
from bgez import logger

__all__ = [
    'BaseSubject', 'Subject',
    'BaseEventEmitter', 'EventEmitter',
]

class BaseSubject(Abstract):

    @abstractmethod
    def subscribe(self, callback: Callable, once: bool = True) -> None:
        '''Subscribes `callback` to this subject.'''

    @abstractmethod
    def notify(self, *args: tuple, catch: bool = True) -> int:
        '''Notifies the `callbacks` for this subject.'''

    @abstractmethod
    def unsubscribe(self, handle: Any) -> None:
        '''Unsubscribes `callback` from this subject.'''

    @property
    @abstractmethod
    def subscribers(self) -> Iterable:
        '''Returns the list of subscribers for this subject.'''

    @abstractmethod
    def clear(self) -> None:
        '''Clears every callback from this subject.'''

    @abstractmethod
    def __len__(self) -> int:
        '''Returns the number of callbacks registered to this subject.'''

    @abstractmethod
    def __contains__(self, handle: Any) -> bool:
        '''Returns if the handle is registered to this subject.'''

class BaseEventEmitter(Abstract):

    @property
    @abstractmethod
    def events(self) -> Iterable[str]:
        '''Returns the current list of available events.'''

    @abstractmethod
    def subject(self, subject: Hashable) -> BaseSubject:
        '''Returns the given `subject`.'''

    @abstractmethod
    def any(self, callback: Callable) -> Callable:
        '''Registers `callback` to be called for any emit.'''

    @abstractmethod
    def on(self, subject: Hashable, callback: Callable) -> Callable:
        '''Registers `callback` to be called on `subject`.'''

    @abstractmethod
    def once(self, subject: Hashable, callback: Callable) -> Callable:
        '''Registers `callback` to be called on `subject` only once.'''

    @abstractmethod
    def remove(self, subject: Hashable, handle: Any) -> None:
        '''Unregisters `callback` from `subject`.'''

    @abstractmethod
    def emit(self, subject: Hashable, *args: tuple, catch: bool = True) -> int:
        '''
        Emits event for `subject` with `args`, blocks the thread.
        Returns the number of callbacks that were called.
        '''

    @abstractmethod
    def listeners(self, subject: Hashable) -> Iterable:
        '''Returns the list of listeners for `subject`.'''

    @abstractmethod
    def clear(self, subject: Hashable) -> None:
        '''Clears `subject`.'''

    @abstractmethod
    def clear_all(self) -> None:
        '''Clears every subject.'''

    @abstractmethod
    def __contains__(self, subject: Hashable) -> bool:
        '''Returns if `subject` is a valid event for this emitter.'''

class Subject(BaseSubject):

    def __init__(self):
        self.__callbacks = OrderedDict()

    @decorator
    def subscribe(self, callback: Callable = None, *, once: bool = False) -> None:
        subscriber = callback
        handle = callback
        if once:
            def subscriber(*args, **kargs): # pylint: disable=E0102
                self.unsubscribe(callback)
                return callback(*args, **kargs)
        self.__callbacks[handle] = subscriber

    def unsubscribe(self, handle: Any) -> None:
        self.__callbacks.pop(handle, None)

    def notify(self, *args, catch: bool = True) -> None:
        callbacks = [*self.__callbacks.values()] # copy
        size = len(callbacks)

        if catch: self._try_exec(callbacks, args)
        else: self._exec(callbacks, args)

        return size

    def _try_exec(self, callbacks, args):
        for callback in callbacks:
            try:
                callback(*args)
            except Exception as e:
                logger.exception("", exc_info=e)

    def _exec(self, callbacks, args):
        for callback in callbacks:
            callback(*args)

    @property
    def subscribers(self) -> Iterable:
        return list(self.__callbacks.keys())

    def clear(self) -> None:
        self.__callbacks.clear()

    def __len__(self) -> int:
        return len(self.__callbacks)

    def __contains__(self, handle: Any) -> bool:
        return handle in self.__callbacks

class EventEmitter(BaseEventEmitter):
    __Subject__ = Subject

    def __init__(self, events: Iterable = None):
        self.__any = self.__Subject__()
        self.__subjects = defaultdict(self.__Subject__)
        self.__valid_subjects = set(events) if events else None

    def __check_subject(self, subject: Hashable) -> None:
        if subject not in self:
            raise AttributeError(f'no subject named "{subject}"')

    @property
    def events(self):
        subjects = self.__valid_subjects
        return tuple(self.__subjects.keys() if subjects is None else subjects)

    def subject(self, subject: Hashable) -> BaseSubject:
        self.__check_subject(subject)
        return self.__subjects[subject]

    def any(self, callback: Hashable = None, *, once: bool = False):
        return self.__any.subscribe(callback, once=once)

    def on(self, subject: Hashable, callback: Callable = None, **kargs) -> Any:
        self.__check_subject(subject)
        return self.__subjects[subject].subscribe(callback, **kargs)

    def once(self, subject: Hashable, callback: Callable = None, **kargs) -> Any:
        self.__check_subject(subject)
        return self.on(subject, callback, **kargs, once=True)

    def remove(self, subject: Hashable, handle: Any) -> None:
        self.__check_subject(subject)
        self.__subjects[subject].unsubscribe(handle)

    def emit(self, subject: Hashable, *args, catch: bool = True) -> None:
        self.__check_subject(subject)
        self.__any.notify(subject, *args, catch=catch)
        return self.__subjects[subject].notify(*args)

    def listeners(self, subject: Hashable) -> Iterable:
        self.__check_subject(subject)
        return self.__subjects[subject].subscribers

    def clear(self, subject: object = None):
        return self.__clear_subject(subject)

    def clear_all(self):
        for subject in self.events:
            self.clear(subject)

    def __clear_subject(self, subject: object):
        self.__check_subject(subject)
        if subject in self.__subjects:
            subject = self.__subjects.pop(subject)
            subject.clear()

    def __contains__(self, subject: Hashable) -> bool:
        return (self.__valid_subjects is None) or (subject in self.__valid_subjects)
