from typing import Union, List, Tuple, Callable

from .timer import Timer, TimeoutTimer, IntervalTimer
from .clock import BaseClock

import heapq

__all__ = [
    'Scheduler'
]

class Scheduler:

    def __init__(self, clock: BaseClock):
        self._timers: List[Tuple[Union[int, float], Timer]] = []
        self.clock = clock

    def timeout(self, timeout, callback: Callable, *args, **kargs) -> None:
        '''Execute a callback after a given amount of time.'''
        timer = TimeoutTimer(self.clock.time, timeout, callback, args, kargs)
        heapq.heappush(self._timers, (timer.deadline, timer))

    def interval(self, interval, callback: Callable, *args, **kargs) -> None:
        '''Repeat a callback periodically over time.'''
        timer = IntervalTimer(self.clock.time, interval, callback, args, kargs)
        heapq.heappush(self._timers, (timer.deadline, timer))

    def __call__(self) -> bool:
        '''
        Returns the state of the PriorityQueue, True if empty, else False.
        '''
        timers = self._timers

        while timers and timers[0][0] < self.clock.time():
            timer = timers[0][1]
            deadline, timer = timer()
            if deadline and callable(timer):
                heapq.heapreplace(timers, (deadline, timer))
            else:
                heapq.heappop(timers)

        return not timers
