from typing import Any

from bgez.core.events import EventEmitter
from bgez.core.time import Scheduler, TickClock

from bge.types import KX_Scene, KX_Camera # pylint: disable=E0401

__all__ = [
    'Scene',
]

class Scene:
    '''
    This is a wrapper around `KX_Scene` objects.
    Acts as an emitter for most events triggered by a BGE Scene.
    `KX_Scene` objects should be wrapped only once.
    '''

    __events__ = 'logic', 'pre_draw_setup', 'pre_draw', 'post_draw', 'remove'

    @classmethod
    def _is_wrapped(self, kx_scene: KX_Scene) -> bool:
        return isinstance(kx_scene.get(Scene, None), Scene)

    def __init__(self, kx_scene: KX_Scene, master=False):
        if self._is_wrapped(kx_scene):
            raise RuntimeError(f'{kx_scene.name} is already wrapped')
        self.__events = EventEmitter(self.__events__)
        self._kx_scene = kx_scene
        self.__master = master
        self.__hook(kx_scene)
        kx_scene[Scene] = self

    def __getattr__(self, attribute):
        return getattr(self._kx_scene, attribute)

    def __check_master(self):
        if self.__master:
            raise RuntimeError(f'cannot end master scene {self.name}')

    def __hook(self, kx_scene: KX_Scene):

        def runner(scheduler):
            def listener(*args):
                scheduler.clock.tick()
                scheduler()
            return listener

        logic_clock = TickClock()
        self.logic = Scheduler(logic_clock)

        pre_draw_setup_clock = TickClock()
        pre_draw_setup = Scheduler(pre_draw_setup_clock)
        self.on('pre_draw_setup', runner(pre_draw_setup))

        pre_draw_clock = TickClock()
        pre_draw = Scheduler(pre_draw_clock)
        self.on('pre_draw', runner(pre_draw))

        post_draw_clock = TickClock()
        post_draw = Scheduler(post_draw_clock)
        self.on('post_draw', runner(post_draw))

        kx_scene.pre_draw_setup.extend((runner(pre_draw_setup), self.__emit('pre_draw_setup')))
        kx_scene.pre_draw.extend((runner(pre_draw), self.__emit('pre_draw')))
        kx_scene.post_draw.extend((runner(post_draw), self.__emit('post_draw')))

        kx_scene.onRemove.append(self.__emit('remove'))

    def __emit(self, subject):
        return lambda *args: self.__events.emit(subject, *args)

    def _update(self):
        self.logic.clock.tick()
        empty = self.logic()
        self.__events.emit('logic')
        return empty

    @property
    def name(self) -> str:
        return self._kx_scene.name

    @name.setter
    def name(self, value):
        self._kx_scene.name = value

    @property
    def objects(self):
        return self._kx_scene.objects

    @property
    def inactives(self):
        return self._kx_scene.objectsInactive

    @property
    def camera(self):
        return self._kx_scene.active_camera

    @camera.setter
    def camera(self, camera: KX_Camera):
        self._kx_scene.active_camera = camera

    @property
    def cameras(self) -> list:
        return self._kx_scene.cameras

    @property
    def lights(self) -> list:
        return self._kx_scene.lights

    @property
    def texts(self) -> list:
        return self._kx_scene.texts

    @property
    def world(self):
        return self._kx_scene.world

    @property
    def filters(self):
        return self._kx_scene.filterManager

    def end(self):
        self.__check_master()
        self._kx_scene.end()

    def replace(self, scene):
        self.__check_master()
        self._kx_scene.replace(scene)

    def suspend(self):
        self._kx_scene.suspend()

    def resume(self):
        self._kx_scene.resume()

    def on(self, *args, **kargs):
        return self.__events.on(*args, **kargs)

    def once(self, *args, **kargs):
        return self.__events.once(*args, **kargs)

    def remove(self, *args):
        return self.__events.remove(*args)
