from typing import Dict, Iterable

from collections import OrderedDict

from bgez.logic.scene import Scene
from bgez.asyn import Promise, Deferred
from bgez import loop

from bge import logic # pylint: disable=E0401

__all__ = [
    'SceneManager',
]

class SceneManager:
    '''
    Manages Scenes.
    - Keeps a track of all alive scenes and wraps them on the go.
    - Responsible for dispatching the logic ticks to callbacks for that event.
    '''

    def __init__(self):
        kx_scene = logic.getCurrentScene()
        self.__scenes = OrderedDict([
            (kx_scene, Scene(kx_scene))
        ])
        self.__waiting: Dict[str, Deferred] = dict()

    def __str__(self):
        scenes = [kx_scene.name for kx_scene in self.__scenes.keys()]
        return f'{self.__class__.__name__} {scenes}'

    def __getitem__(self, item) -> Scene:
        return self.__scenes[logic.getSceneList()[item]]

    def __check_inactive(self, name):
        inactives = self.inactives
        if name not in inactives:
            raise KeyError(f'{name} is not in {inactives}')

    def __addScene(self, name: str, overlay: int) -> Promise:
        self.__check_inactive(name)
        queue = self.__waiting
        if name in queue:
            return queue[name].promise

        queue[name] = deferredScene = Deferred()
        logic.addScene(name, overlay) # actual scene adding
        return deferredScene.promise

    def _update(self) -> bool:
        waiting = self.__waiting
        scenes = self.__scenes
        empty = True

        for kx_scene in logic.getSceneList():
            if kx_scene not in scenes:

                # wrap the BGE scene
                scenes[kx_scene] = scene = Scene(kx_scene)

                # resolve waiting promises
                name = kx_scene.name
                if name in waiting:
                    waiting.pop(name).resolve(scene)

        # dispatch component logic ticks
        for scene in self.__scenes.values():
            if not scene._update() and empty:
                empty = False

        return empty

    @property
    def current(self) -> Scene:
        '''Returns the current scene.'''
        return self.__scenes[logic.getCurrentScene()]

    @property
    def inactives(self) -> Iterable[str]:
        '''Returns the names of the inactive scenes'''
        return logic.getInactiveSceneList()

    def overlay(self, name: str) -> Promise:
        return self.__addScene(name, 1)

    def underlay(self, name: str) -> Promise:
        return self.__addScene(name, 0)
