'''
Global registry for the application event loop.
'''
from typing import Any, Callable

import inspect

from functools import wraps

from bgez.asyn import iscoroutine

__LOOP__ = None

__all__ = [
    'get',
    'set',
    'task',
]

def get():
    '''Get the global eventloop.'''
    if __LOOP__ is None:
        raise RuntimeError('no eventloop is set...')
    return __LOOP__

def set(loop):
    '''Set the global eventloop, old state will be lost!'''
    global __LOOP__
    __LOOP__ = loop

def task(function) -> Callable:
    '''
    Converts a function/task to run in the global eventloop.

    You can decorate both async and sync methods. Will always return a promise after decoration.
    '''
    if getattr(function, '__task__', False):
        return function

    @wraps(function)
    def wrapper(*args, **kargs):
        value = function(*args, **kargs)
        loop = get()

        if iscoroutine(value):
            return loop.register(value)

        async def coroutine_factory():
            return value
        return loop.register(coroutine_factory())

    wrapper.__task__ = True
    return wrapper

async def value(v):
    '''Convert a potentially non-awaitable value to be used with `await`.'''
    return (await v) if inspect.isawaitable(v) else v
