import json

from bgez.logic import Scene
from bgez.logic.input import Keymap
from bgez.network import UDPClient, NetworkManager

from bgez.blender import game
from mathutils import Vector

from game import SERVER

CLIENT = ('0.0.0.0', 0)
KEYMAP = Keymap()

# For experiments purpose, hook the network manager at this point.
MANAGER = NetworkManager()
game.use(MANAGER)

@game.on('start')
class Client:
    '''
    Again, no need to define classes like that. Did it for fun.
    Turns out its a useful pattern.

    Usually you would just hook a function to this `game.on('start')` event,
    but when using class instances you will be able to store some state inside
    the instance, which is cool.

    This comment is useless, right?
    '''

    # Store some inputs to later move the client avatar.
    w = KEYMAP.input('w')
    a = KEYMAP.input('a')
    s = KEYMAP.input('s')
    d = KEYMAP.input('d')
    shift = KEYMAP.input('leftshift')
    space = KEYMAP.input('space')

    def __init__(self, scene: Scene):
        '''
        Spawn the client avatar right away.
        Setup the player registry (dict, to keep track of others).
        Setup the UDP client that will send messages to the server.
        Setup the UDP client to send the position on pre_draw (physics applied)
        Setup the logic to run every frame.
        '''
        self.cube = scene.addObject('MyCube')
        self.scene = scene
        self.players = {}

        self.client = UDPClient(MANAGER, host = CLIENT, buffer_size = 2 << 16)
        self.client.connect(SERVER)
        self.client.on('read', self._on_message)

        scene.on('pre_draw_setup', self._on_pre_draw_setup)
        scene.logic.interval(1, self._user_input)

        scene.camera = scene.cameras['MyCamera']
        scene.camera.setParent(self.cube)

    def _on_pre_draw_setup(self, _):
        '''
        We want to send the position on pre_draw_setup because it is the
        closest phase to the physics step. In logic the position wouldn't
        have been updated after setting the position.
        '''
        self.client.send(json.dumps(
            list(self.cube.worldPosition)
        ).encode())

    def _on_message(self, client, message):
        '''
        When we receive a message from the server.
        Interpret the message as { some_id: [x, y, z] }
        Move every other player avatar.
        '''
        data = json.loads(message.decode())
        for id, position in data.items():
            if id not in self.players: # Didn't know player yet, creating it.
                self.players[id] = self.scene.addObject('MyCube')
            self.players[id].worldPosition = position

    def _user_input(self):
        '''
        Get the current velocity and add some based on user input.
        '''
        velocity = self.cube.getLinearVelocity()
        movement = Vector((0, 0, 0))

        if self.w.pressed: movement.y += 1
        if self.a.pressed: movement.x -= 1
        if self.s.pressed: movement.y -= 1
        if self.d.pressed: movement.x += 1
        movement.normalize()

        if self.shift.pressed:
            movement *= 3

        if self.space.pressing:
            movement.z += 1

        movement *= 5

        self.cube.setLinearVelocity((movement.x, movement.y, velocity.z + movement.z))
