import json
import time

from bgez.blender import game
from bgez.logic import Scene

from bgez.network import NetworkManager, UDPServer

from game import SERVER

# For experiments purpose, hook the network manager at this point.
manager = NetworkManager()
game.use(manager)

@game.on('start')
class Server:
    '''
    Usually we use a callback function to respond to `game.on('start')` events,
    but I figure that if we were to decorate a class, it would create an
    instance when the event is fired, and from there we can store some internal
    state more easily.

    Basically, you don't have to hook a class to the event, but you can.
    Any callable will do (functions mostly).
    '''

    TIMEOUT = 3

    def __init__(self, scene: Scene):
        '''
        This is called when `start` is emitted.
        Setup the player registry (dict), and last message times.
        Setup the UDP server that will listen to the messages.
        Setup the logic to run every frame.
        '''
        self.players = {}
        self.times = {}

        self.scene = scene

        self.server = UDPServer(manager, SERVER)
        self.server.on('connect', self._on_client)
        self.server.start()

        scene.logic.interval(1, self._update_clients)

    def addPlayer(self, client):
        '''
        Add a player in the scene and track it.
        '''
        self.updatePlayer(client)
        cube = self.scene.addObject('MyCube')
        self.players[client] = cube
        print(f'Connected: {client}')
        return cube

    def validatePlayer(self, client):
        '''
        Check if a client still has some time before timing-out.
        '''
        return self.times[client] + self.TIMEOUT > time.monotonic()

    def updatePlayer(self, client):
        '''
        Update the last time a client sent a message.
        (Keep alive).
        '''
        self.times[client] = time.monotonic()

    def removePlayer(self, client):
        '''
        Remove a player from the scene as well as from the registries.
        '''
        self.players.pop(client).endObject()
        self.times.pop(client)
        print(f'Timed-out: {client}')
        client.close()

    def _on_client(self, client):
        '''
        When we receive a new message from someone.
        '''
        cube = self.addPlayer(client) # Track the player.

        def _on_message(client, message):
            '''
            When we receive a message from that client,
            read the message as [x, y, z] and assign it
            to its avatar.
            '''
            message = message.decode()
            x, y, z = json.loads(message)
            cube.worldPosition = x, y, z

            self.updatePlayer(client) # Update timeout.

        client.on('read', _on_message)

    def _update_clients(self):
        '''
        Sends the position of every players to every clients.

        If we did not receive a message from a client for a short period of time,
        consider this client disconnected (UDP doesn't handle connections).
        '''
        positions = {}

        for client, player in list(self.players.items()):
            if not self.validatePlayer(client):
                self.removePlayer(client) # Client disconnected.
                continue

            positions[id(player)] = list(player.worldPosition)

        for client, player in self.players.items():
            data = positions.copy()
            data.pop(id(player))
            client.send(json.dumps(data).encode())
